use std::time::Duration;

use color_eyre::{Result, eyre::Ok};
use crossterm::event::{Event, KeyCode, KeyEventKind, MouseEventKind};
use rand::seq::SliceRandom;
use ratatui::{
    Terminal,
    layout::{Constraint, Layout},
    prelude::CrosstermBackend,
    style::Stylize,
};
use tokio::{
    sync::mpsc::{self, UnboundedReceiver, UnboundedSender},
    task::JoinHandle,
    time,
};
pub enum Message {
    Quit,
    Tick,
    Render,
    StartSimulation,
    SwapEvent(usize, usize),
    SetEvent(usize, u32),
}

pub struct Tui {
    terminal: Terminal<CrosstermBackend<std::io::Stdout>>,
    tick_rate: f64,
    frame_rate: f64,
    event_rx: UnboundedReceiver<Message>,
    event_tx: UnboundedSender<Message>,
    sorting_tx: UnboundedSender<SortingMessage>,
    should_exit: bool,
    vec: Vec<u32>,
}
use crate::{
    drawableVec::{self, DrawableVec, SortingMessage},
    graphics::vertical_barchart,
};

#[derive(Clone, Debug)]
pub enum UpdateCommand {
    None,
    Quit,
}

impl Tui {
    pub fn new() -> Result<Self> {
        let terminal = ratatui::init();
        let (event_tx, event_rx): (UnboundedSender<Message>, UnboundedReceiver<Message>) =
            mpsc::unbounded_channel();
        let (sorting_tx, sorting_rx) = mpsc::unbounded_channel();
        let mut vec = (Vec::from_iter(1..100));
        vec.shuffle(&mut rand::rng());
        let drawableVec = DrawableVec::new(event_tx.clone(), sorting_rx, vec.clone());

        let sorter = tokio::spawn(drawableVec.run());

        Ok(Self {
            terminal,
            tick_rate: 20.,
            frame_rate: 120.,
            event_rx,
            event_tx,
            sorting_tx,
            should_exit: false,
            vec: vec,
        })
    }

    pub async fn run(&mut self) -> Result<()> {
        let tick_rate = Duration::from_secs_f64(1.0 / self.tick_rate);
        let frame_rate = Duration::from_secs_f64(1.0 / self.frame_rate);
        let mut tick_interval = time::interval(tick_rate);
        let mut frame_interval = time::interval(frame_rate);
        while !self.should_exit {
            tokio::select! {
                _tick = tick_interval.tick() => {
                    self.event_tx.send(Message::Tick)?;
                    self.sorting_tx.send(SortingMessage::Tick)?
                }
                _frame = frame_interval.tick() => {
                    self.event_tx.send(Message::Render)?
                }
                Some(message) = self.event_rx.recv() => {
                    match self.update(message).await? {
                        UpdateCommand::Quit => return {
                            self.should_exit = true;
                            Ok(())
                        },
                        UpdateCommand::None => continue,
                    }
                }
                Result::Ok(ready) = tokio::task::spawn_blocking(|| crossterm::event::poll(Duration::from_millis(100))) => {
                    match ready? {
                        true => {
                            let event = crossterm::event::read()?;
                            self.handle_event(event)?;
                        }
                        false => continue,
                    }
                }
            }
        }
        Ok(())
    }

    async fn update(&mut self, message: Message) -> Result<UpdateCommand> {
        match message {
            Message::Quit => Ok(UpdateCommand::Quit),
            Message::Tick => Ok(UpdateCommand::None),
            Message::Render => {
                self.view()?;
                Ok(UpdateCommand::None)
            }
            Message::StartSimulation => todo!(),
            Message::SwapEvent(a, b) => {
                self.vec.swap(a, b);
                Ok(UpdateCommand::None)
            }
            Message::SetEvent(i, e) => {
                self.vec.insert(i, e);
                Ok(UpdateCommand::None)
            }
        }
    }

    fn view(&mut self) -> Result<()> {
        self.terminal.draw(|frame| {
            let [title, vertical, horizontal] = Layout::vertical([
                Constraint::Length(1),
                ratatui::layout::Constraint::Fill(1),
                Constraint::Fill(1),
            ])
            .spacing(1)
            .areas(frame.area());
            frame.render_widget("Sorting!".bold().into_centered_line(), title);
            frame.render_widget(vertical_barchart(&self.vec), horizontal);
        })?;

        Ok(())
    }

    fn handle_event(&self, event: Event) -> Result<()> {
        match event {
            Event::Key(key) => {
                if key.kind == KeyEventKind::Press && key.code == KeyCode::Esc {
                    self.event_tx.send(Message::Quit)?;
                } else if key.kind == KeyEventKind::Press && key.code == KeyCode::Enter {
                    self.event_tx.send(Message::StartSimulation)?;
                }
            }
            _ => {}
        }
        Ok(())
    }
}
